#include "header\gba.h"
#include "header\bg.h"
#include "header\timer.h"
#include "header\dma.h"
#include "header\irq.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\blend.h"

#include "header\font\BGFont.h"
//#include "header\font\BitMapFont.h"

#include "lib.h"
#include "syscall.h"
#include "resource.h"

#include <stdlib.h>

//Mode0
//BG0 
//BG2 
//BG3 Ֆ

//---------------------------------------------------------------------------
//is
#define STATE_INIT 				1
#define STATE_RESET				2
#define STATE_PLAYER			3
#define STATE_COMPUTER			4
#define STATE_CLEAR_CHK			5
#define STATE_CLEAR				6

//Ֆ
#define BOARD_LENGTH_CX			10
#define BOARD_LENGTH_CY			10
#define BOARD_NONE				1
#define BOARD_WALL				2
#define BOARD_PLAYER			3
#define BOARD_COMPUTER			4

//\pBG^C
#define BG_TILE_WALL			0
#define BG_TILE_EMPTY			2
#define BG_TILE_CELL			4
#define BG_TILE_STONE_RED		6
#define BG_TILE_STONE_BLUE		8

//SRAMZ[ü
#define SRAM_RANDOM_OFFSET		0

//---------------------------------------------------------------------------
//Ֆ
typedef struct {
	u8  cell[BOARD_LENGTH_CX][BOARD_LENGTH_CY];
	u8  turn;		//
	u8  turnCnt;
} ST_BOARD;

//J[\
typedef struct {
	s16 x;
	s16 y;
	u16 cx;
	u16 cy;
	s16 moveX;
	s16 moveY;
	s16 minX;		//ړ
	s16 minY;
	s16 maxX;
	s16 maxY;
	u16 sprNo;
	u16 sprTile;
	u16 sprSize;
	u16 sprForm;
	u8  key;		//ړp
} ST_CURSOR;

//---------------------------------------------------------------------------
//lib.c
extern ST_TIMER Timer;

//---------------------------------------------------------------------------
ST_BG     BG0;
ST_BG     BG2;
ST_BG     BG3;
ST_CURSOR Cursor;
ST_BOARD  Board;

u8  State;
u32 Random;
//---------------------------------------------------------------------------
void StateInit();
void StateReset();
void StatePlayer();  
void StateComputer();
void StateClearChk();
void StateClear();

void InitWS();
void InitBG();
void InitSpr();
void InitKey();
void IRQUserHandler();

void InitCursor();
void InitStage();

void ProcKeyInput();
bool IsMoveCursor(ST_CURSOR* p);
void MoveCursor(ST_CURSOR* p);
void DrawSprCursor(ST_CURSOR* p);

bool IsPutStone(u8 x, u8 y);
void PutStone(u8 x, u8 y);
void DrawStone(u8 x, u8 y);

bool IsGameClear();
bool Is5Ren();
u8   GetRenCnt(u8 x, u8 y);
u8   GetRenCntSub(u8 argX, u8 argY, s8 StepX, s8 StepY);

u8   ComGetRandomPoint();
u8   ComGetChoicePoint();
u16  ComGetChoicePointSub(u8 x, u8 y);
u16  ComGetChoicePointSubSub(u8 argX, u8 argY, s8 StepX, s8 StepY);

void DrawBG(u8 x, u8 y, u8 tile, u8 map, u8 col);
//---------------------------------------------------------------------------
int main()
{
	State = STATE_INIT;

	for(;;)
	{
		WaitForVsync();

		switch(State)
		{
			case STATE_INIT:      StateInit();     break;
			case STATE_RESET:     StateReset();    break;
			case STATE_PLAYER:    StatePlayer();   break;
			case STATE_COMPUTER:  StateComputer(); break;
			case STATE_CLEAR_CHK: StateClearChk(); break;
			case STATE_CLEAR:     StateClear();    break;
			default:
				DebugMessage((u8*)"Err: Main State");
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
void StateInit()
{
	InitWS();
	InitBG();
	InitSpr();
	InitKey();

	SetMode(MODE_0 | BG0_ENABLE | BG2_ENABLE | BG3_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);
	State = STATE_RESET;
}
//---------------------------------------------------------------------------
void InitWS()
{
	REG_WSCNT = CST_ROM0_1ST_3WAIT | CST_ROM0_2ND_1WAIT;
}
//---------------------------------------------------------------------------
void InitBG()
{
	//BG0 
	//BG2 
	//BG3 Ֆ

	BG0.chr = 1;
	BG0.map = 0;
	BG0.col = 0;

	BG2.chr = 2;	//BG3Ɠ
	BG2.map = 2;
	BG2.col = 1;	//BG3Ɠ

	BG3.chr = 2;
	BG3.map = 1;
	BG3.col = 1;

	REG_BG0CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG0.chr) | BG_MAPBASE(BG0.map) | 0);

	REG_BG2CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG2.chr) | BG_MAPBASE(BG2.map) | 0);

	REG_BG3CNT = (BG_SIZEA_256_256     | BG_COLOR_16         |
				  BG_CHARBASE(BG3.chr) | BG_MAPBASE(BG3.map) | 0);

	BGSetPal16  (BG0.col, (u16*)&BGFont_palette);
	BGSetChrData(BG0.chr, (u16*)&BGFont_data, BGFont_DATA_SIZE);

	BGSetPal16  (BG3.col, (u16*)&game_bg_pal_bin);
	BGSetChrData(BG3.chr, (u16*)&game_bg_img_bin, GAME_BG_IMG_SIZE);
}
//---------------------------------------------------------------------------
void InitSpr()
{
	SpriteInit();

	SpriteSetPal16((u16*)&game_spr_pal_bin);
	SpriteSetData ((u16*)&game_spr_img_bin, GAME_SPR_IMG_SIZE);
}
//---------------------------------------------------------------------------
void InitKey()
{
	KeyInit();

	//L[͗p̃^C}ݒ
	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER0;
	REG_IME      = IRQ_MASTER_ON;

	REG_TM0D     = 65536 - 2621; // (((16*1024*1024) / 64) * 10) / 1000
	REG_TM0CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;
}
//---------------------------------------------------------------------------
void IRQUserHandler()
{
	Timer.clockTick += TIMER_INTMS;
	REG_IF = IRQ_BIT_TIMER0;
}
//---------------------------------------------------------------------------
void InitCursor()
{
	ST_CURSOR* p = &Cursor;

	p->x       = 1;
	p->y       = 1;
	p->cx      = 16;
	p->cy      = 16;
	p->moveX   = 1;
	p->moveY   = 1;
	p->minX    = 1;
	p->minY    = 1;
	p->maxX    = BOARD_LENGTH_CX - 1;
	p->maxY    = BOARD_LENGTH_CY - 1;
	p->sprNo   = 0;
	p->sprTile = 0;
	p->sprSize = SP_SIZE_16;
	p->sprForm = SP_SQUARE;
	p->key     = 0x0000;

	SpriteSetSize(p->sprNo, p->sprSize, p->sprForm, SP_COLOR_16);
	SpriteSetChr (p->sprNo, p->sprTile);

	DrawSprCursor(p);
}
//---------------------------------------------------------------------------
void StateReset()
{
	InitCursor();
	InitStage();

	State = STATE_PLAYER;
}
//---------------------------------------------------------------------------
void StatePlayer()
{
	ProcKeyInput();
}
//---------------------------------------------------------------------------
void ProcKeyInput()
{
	rand();							//Q[Jnɗ𔭐

	if( KeyWaitTimer() == TRUE )
	{
		return;
	}
	if( KeyRepeat() == TRUE )		//keyGet֐ĂԑOɕKsȂ
	{
		return;
	}

	u8 key = KeyGet();
	if(key == 0x00)
	{
		return;
	}

	ST_CURSOR* p = &Cursor;

	p->key = 0x00;
	if(key & KEY_UP   ) p->key |= KEY_UP;
	if(key & KEY_DOWN ) p->key |= KEY_DOWN;
	if(key & KEY_LEFT ) p->key |= KEY_LEFT;
	if(key & KEY_RIGHT) p->key |= KEY_RIGHT;
	if(key & KEY_A    ) p->key |= KEY_A;

	if( IsMoveCursor(p) == TRUE )
	{
		MoveCursor(p);
		DrawSprCursor(p);
	}

	if(p->key & KEY_A)
	{
		if( IsPutStone(p->x, p->y) == TRUE )
		{
			PutStone (p->x, p->y);
			DrawStone(p->x, p->y);
			State = STATE_CLEAR_CHK;
		}
	}
}
//---------------------------------------------------------------------------
void StateComputer()
{
	u16 choicePoint;
	u8  x, y;

	//߂l: u8Pʁuxxxx yyyyv
	choicePoint = ComGetChoicePoint();

	//IȂꍇ
	if(choicePoint == 0x0000)
	{
		choicePoint = ComGetRandomPoint();
	}

	x = (choicePoint >> 4) & 0x0f;
	y = (choicePoint     ) & 0x0f;

	PutStone (x, y);
	DrawStone(x, y);
	State = STATE_CLEAR_CHK;
}
//---------------------------------------------------------------------------
//߂l: u8Pʁuxxxx yyyyv
u8 ComGetChoicePoint()
{
	u16 retPoint, tmpPoint;
	s16 x,y;

	retPoint = 0x0000;
	for(y=1; y<BOARD_LENGTH_CY-1; y++)
	{
		for(x=1; x<BOARD_LENGTH_CX-1; x++)
		{
			//߂l: u16Pʁupppp pppp xxxx yyyyvp͗Dx
			tmpPoint = ComGetChoicePointSub(x, y);

			//Dx̍Wi[܂
			if((tmpPoint & 0xff00) > (retPoint & 0xff00))
			{
				retPoint = tmpPoint;
			}
		}
	}

	return retPoint & 0x00ff;
}
//---------------------------------------------------------------------------
u16 ComGetChoicePointSub(u8 x, u8 y)
{
	if(Board.cell[x][y] == BOARD_NONE)
	{
		return 0x0000;
	}

	u16 point[4];

	point[0] = ComGetChoicePointSubSub(x, y, 1, 0);	//E
	point[1] = ComGetChoicePointSubSub(x, y, 0, 1);	//
	point[2] = ComGetChoicePointSubSub(x, y, 1, 1);	//E
	point[3] = ComGetChoicePointSubSub(x, y, 1,-1);	//E

	u16 bestPoint = point[0];

	if(bestPoint < point[1]) bestPoint = point[1];
	if(bestPoint < point[2]) bestPoint = point[2];
	if(bestPoint < point[3]) bestPoint = point[3];

/*
	if(Board.cell[x][y] == BOARD_PLAYER)
	{
		u8 tmpPriority = (bestPoint & 0xff00) >> 8;

		//vC̘AQ菬ꍇ͖
		if(tmpPriority <= 2)
		{
			//DxŒɂ܂
			bestPoint = 0x0000;
		}
	}
*/
	return bestPoint;
}
//---------------------------------------------------------------------------
u16 ComGetChoicePointSubSub(u8 argX, u8 argY, s8 StepX, s8 StepY)
{
	u8   x        = argX;
	u8   y        = argY;
	u8   stone    = Board.cell[x][y];
	u8   renCnt   = 0;

	bool emptyFlg = FALSE;
	u8   emptyX   = 0;
	u8   emptyY   = 0;

	do {
		x += StepX;
		y += StepY;
		renCnt++;

		if(Board.cell[x][y] == BOARD_NONE)
		{
			if(emptyFlg == FALSE)
			{
				emptyFlg = TRUE;
				emptyX   = x;
				emptyY   = y;
				continue;
			}
			else
			{
				break;
			}
		}

	} while(Board.cell[x][y] == stone);

	if(emptyFlg == TRUE)
	{
		return (renCnt << 8) | (emptyX << 4) | (emptyY);
	}

	//FALSE
	return 0x0000;
}
//---------------------------------------------------------------------------
void StateClearChk()
{
	if( IsGameClear() == TRUE )
	{
		State = STATE_CLEAR;
		return;
	}

	if(Board.turn == BOARD_PLAYER)
	{
		Board.turn = BOARD_COMPUTER;
		State      = STATE_COMPUTER;
	}
	else
	{
		Board.turn = BOARD_PLAYER;
		State      = STATE_PLAYER;
	}

	Board.turnCnt++;
}
//---------------------------------------------------------------------------
void StateClear()
{
	if(Board.turnCnt >= BOARD_LENGTH_CX * BOARD_LENGTH_CY)
	{
		BGDrawStr(BG0.map,  0, 19, (u8*)"Draw");
	}

	if(Board.turn == BOARD_PLAYER)
	{
		BGDrawStr(BG0.map,  0, 19, (u8*)"Player Win!");
	}
	else
	{
		BGDrawStr(BG0.map,  0, 19, (u8*)"Computer Win!");
	}


	for(;;){}
}
//---------------------------------------------------------------------------
void InitStage()
{
	s16 x,y;

	//Ֆʂ̏
	ST_BOARD* p = &Board;
	for(y=0; y<BOARD_LENGTH_CY; y++)
	{
		for(x=0; x<BOARD_LENGTH_CX; x++)
		{
			p->cell[x][y] = BOARD_NONE;
		}
	}

	for(x=0; x<BOARD_LENGTH_CX; x++)
	{
		p->cell[x][0]                 = BOARD_WALL;
		p->cell[x][BOARD_LENGTH_CY-1] = BOARD_WALL;
	}
	for(y=0; y<BOARD_LENGTH_CY; y++)
	{
		p->cell[0][y]                 = BOARD_WALL;
		p->cell[BOARD_LENGTH_CX-1][y] = BOARD_WALL;
	}

	//ڂ̕`
	for(y=1; y<BOARD_LENGTH_CY-1; y++)
	{
		for(x=1; x<BOARD_LENGTH_CX-1; x++)
		{
			DrawBG(x*2, y*2, BG_TILE_CELL, BG3.map, BG3.col);
		}
	}

	p->turn    = BOARD_PLAYER;
	p->turnCnt = 0;
}
//---------------------------------------------------------------------------
void DrawBG(u8 x, u8 y, u8 tile, u8 map, u8 col)
{
	BGDrawChar(map, x+0, y+0, col, tile+0 );
	BGDrawChar(map, x+1, y+0, col, tile+1 );
	BGDrawChar(map, x+0, y+1, col, tile+32);
	BGDrawChar(map, x+1, y+1, col, tile+33);
}
//---------------------------------------------------------------------------
bool IsMoveCursor(ST_CURSOR* p)
{
	s16 tx1 = p->x;
	s16 ty1 = p->y;

	if(p->key & KEY_UP   ) ty1 -= p->moveY;
	if(p->key & KEY_DOWN ) ty1 += p->moveY;
	if(p->key & KEY_LEFT ) tx1 -= p->moveX;
	if(p->key & KEY_RIGHT) tx1 += p->moveX;

	if(tx1< p->minX || ty1< p->minY) return FALSE;
	if(tx1>=p->maxX || ty1>=p->maxY) return FALSE;

	return TRUE;
}
//---------------------------------------------------------------------------
void MoveCursor(ST_CURSOR* p)
{
	if(p->key & KEY_UP   ) p->y-=p->moveY;
	if(p->key & KEY_DOWN ) p->y+=p->moveY;
	if(p->key & KEY_LEFT ) p->x-=p->moveX;
	if(p->key & KEY_RIGHT) p->x+=p->moveX;
}
//---------------------------------------------------------------------------
void DrawSprCursor(ST_CURSOR* p)
{
	u8 x = p->x * p->cx;
	u8 y = p->y * p->cy;
	SpriteMove(p->sprNo, x, y);
}
//---------------------------------------------------------------------------
bool IsPutStone(u8 x, u8 y)
{
	return Board.cell[x][y] == BOARD_NONE ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
bool IsGameClear()
{
	if( Is5Ren() == TRUE )
	{
		return TRUE;
	}

	if(Board.turnCnt >= BOARD_LENGTH_CX * BOARD_LENGTH_CY)
	{
		return TRUE;
	}

	return FALSE;
}
//---------------------------------------------------------------------------
void DrawStone(u8 x, u8 y)
{
	u8 bgTile;

	if(Board.turn == BOARD_PLAYER) bgTile = BG_TILE_STONE_BLUE;
	else                           bgTile = BG_TILE_STONE_RED;

	DrawBG(x*2, y*2, bgTile, BG2.map, BG2.col);
}
//---------------------------------------------------------------------------
bool Is5Ren()
{
	s16 x,y;

	for(y=1; y<BOARD_LENGTH_CY-1; y++)
	{
		for(x=1; x<BOARD_LENGTH_CX-1; x++)
		{
			if( GetRenCnt(x, y) >= 5 )
			{
				return TRUE;
			}
		}
	}

	return FALSE;
}
//---------------------------------------------------------------------------
u8 GetRenCnt(u8 x, u8 y)
{
	if(Board.cell[x][y] == BOARD_NONE)
	{
		return 0;
	}

	u8 cnt[4];

	cnt[0] = GetRenCntSub(x, y, 1, 0);	//E
	cnt[1] = GetRenCntSub(x, y, 0, 1);	//
	cnt[2] = GetRenCntSub(x, y, 1, 1);	//E
	cnt[3] = GetRenCntSub(x, y, 1,-1);	//E

	u8 maxCnt = cnt[0];

	if(maxCnt < cnt[1]) maxCnt = cnt[1];
	if(maxCnt < cnt[2]) maxCnt = cnt[2];
	if(maxCnt < cnt[3]) maxCnt = cnt[3];

	return maxCnt;
}
//---------------------------------------------------------------------------
u8 GetRenCntSub(u8 argX, u8 argY, s8 StepX, s8 StepY)
{
	u8 x     = argX;
	u8 y     = argY;
	u8 stone = Board.cell[x][y];
	u8 cnt   = 0;

	do {
		cnt++;
		x += StepX;
		y += StepY;

	} while(Board.cell[x][y] == stone);

	return cnt;
}
//---------------------------------------------------------------------------
void PutStone(u8 x, u8 y)
{
	Board.cell[x][y] = Board.turn;
}
//---------------------------------------------------------------------------
u8 ComGetRandomPoint()
{
	u8 x;
	u8 y;

	do{
		x = (rand() % (BOARD_LENGTH_CX - 2)) + 1;
		y = (rand() % (BOARD_LENGTH_CY - 2)) + 1;

	} while(Board.cell[x][y] != BOARD_NONE);

	return (x<<4) | (y);
}
